
import sys
from io import StringIO
import contextlib

@contextlib.contextmanager
def stdoutIO(stdout=None):
    old = sys.stdout
    if stdout is None:
        stdout = StringIO()
    sys.stdout = stdout
    yield stdout
    sys.stdout = old

code = """
i = [0,1,2]
for j in i :
    print(j)
"""
with stdoutIO() as s:
    try:
        exec(code)
    except:
        print("Something wrong with the code")
print("out:", s.getvalue())


import ast, re, sys, traceback
from pathlib import Path

from 木兰.分析器.词法分析器 import 分词器
from 木兰.分析器.语法分析器 import 语法分析器

运行时木兰路径 = str(Path("site-packages/木兰/"))

报错_列表索引 = "取列表内容时，索引超出范围"
报错_层级 = "调用层级如下"
报错_除零 = "请勿除以零"
报错_递归 = "递归过深。请确认: 1、的确需要递归 2、递归的收敛正确"
报错_按索引取项 = "不支持按索引取项"
报错_无属性 = "没有属性"
参考_enter = "\n参考：https://stackoverflow.com/questions/1984325/explaining-pythons-enter-and-exit"

def 反馈信息(例外, 源码文件=None):
    提神符 = " 😰 "
    if sys.platform == 'win32':
        提神符 = "（>﹏<）"
    return 提神符 + "\n".join(中文化(例外, 源码文件))

def 中文化(例外, 源码文件=None):
    类型 = 例外.__class__.__name__
    原信息 = str(例外)
    exc_type, exc_value, 回溯信息 = sys.exc_info()
    各层 = traceback.extract_tb(回溯信息)
    # print(repr(各层))
    各行 = []

    行信息 = 提取(各层)
    if len(行信息) > 0:
        关键 = 取关键信息(提示(类型, 原信息), 行信息[0])
    各行.append(关键)
    for 行号, 行 in enumerate(行信息, start=1):
        if 源码文件 is None and 行.文件名 == "【标准输入】":
            return [关键 + f"，见第{行.行号}行"]
        else:
            # 在第二层前显示
            if 行号 == 2:
                各行.append(报错_层级)

            各行.append(("见" if 行.文件名 == 源码文件 else f"“{行.文件名}”") + f"第{行.行号}行：{行.内容}")
    return 各行

class 层信息:
    def __init__(self, 行号, 内容, 文件名):
        self.行号, self.内容, self.文件名 = 行号, 内容, 文件名

def 提取(各层):
    各行 = []
    for 层号 in range(len(各层) - 1, -1, -1):
        层 = 各层[层号]
        文件名 = 层.filename
        if 文件名.find(运行时木兰路径) == -1:
            各行.append(层信息(层.lineno, 层.line, 文件名))

    return 各行

def 取关键信息(基本信息, 首行):
    代码行 = 首行.内容
    问题变量 = []
    try:
        if 基本信息.find(报错_按索引取项) > 0:
            节点 = 语法分析器(分词器).分析(代码行)
            问题变量 = 诊断问题(节点)
        匹配 = re.search("(.*)" + 报错_无属性 + "‘(.*)’", 基本信息)
        if 匹配:
            节点 = 语法分析器(分词器).分析(代码行)

            # TODO: 重构以避免重复匹配
            问题属性 = 匹配.group(2)
            问题变量 = 诊断无属性问题(节点, 问题属性)
    except ValueError:
        return 基本信息

    if len(问题变量) == 0:
        return 基本信息
    elif len(问题变量) == 1:
        return f"{基本信息}，看看‘{问题变量[0]}’"
    else:
        return f"{基本信息}，看看{问题变量}"


def 诊断问题(节点):
    所有变量 = []
    if isinstance(节点, list):
        for 子节点 in 节点:
            所有变量 += 诊断问题(子节点)
    elif isinstance(节点, int) or isinstance(节点, str):
        pass
    else:
        if 节点 == None:
            return
        for 属性 in ast.iter_fields(节点):
            if type(节点).__name__ == "Subscript":
                if 属性[0] == "value":
                    return [属性[1].id]
            所有变量 += 诊断问题(属性[1])
    return 所有变量

def 诊断无属性问题(节点, 问题属性):
    所有变量 = []
    if isinstance(节点, list):
        for 子节点 in 节点:
            所有变量 += 诊断无属性问题(子节点, 问题属性)
    elif isinstance(节点, int) or isinstance(节点, str):
        pass
    else:
        if 节点 == None:
            return
        有问题 = False
        对象名 = ""
        for 属性 in ast.iter_fields(节点):
            if type(节点).__name__ == "Attribute":
                # print(属性[0] + ": " + 属性[1])
                if 属性[0] == "value":
                    对象名 = 属性[1].id
                elif 属性[0] == "attr" and 属性[1] == 问题属性:
                    有问题 = True
            所有变量 += 诊断无属性问题(属性[1], 问题属性)
        if 有问题:
            return [对象名]
    return 所有变量

def 提示(类型, 原信息):
    if 类型 == 'NameError':
        return re.sub(r"name '(.*)' is not defined", r"请先定义‘\1’再使用", 原信息)
    elif 类型 == 'ZeroDivisionError':
        return 报错_除零
    elif 类型 == 'RecursionError':
        return 报错_递归
    elif 类型 == 'UnboundLocalError':
        return re.sub(
            r"local variable '(.*)' referenced before assignment",
            r"请先对本地变量‘\1’赋值再引用",
            原信息)
    elif 类型 == 'KeyError':
        return "字典中不存在此键：" + 原信息
    elif 类型 == 'TypeError':
        模式 = 'can only concatenate str \(not "(.*)"\) to str'
        无法取项 = "'(.*)' object is not subscriptable"
        if re.match(模式, 原信息):
            return re.sub(模式, r'字符串只能拼接字符串，请将“\1”先用 str() 转换', 原信息)
        匹配 = re.search(无法取项, 原信息)
        if 匹配:
            return f'{类型中文化(匹配.group(1))}{报错_按索引取项}'
    elif 类型 == 'IndexError' and 原信息 == "list index out of range":
        return 报错_列表索引
    elif 类型 == 'AttributeError':
        信息 = "需要添加此属性：" + 原信息
        if 原信息 == "__enter__":
            信息 += 参考_enter
            return 信息

        无属性 = "'(.*)' object has no attribute '(.*)'"
        匹配 = re.search(无属性, 原信息)
        if 匹配:
            return f'{类型中文化(匹配.group(1))}{报错_无属性}‘{匹配.group(2)}’'
    elif 类型 == 'FileNotFoundError':
        return re.sub(r"\[Errno 2\] No such file or directory: '(.*)'",
            r"没找到文件或路径：‘\1’",
            原信息)
    elif 类型 == 'ModuleNotFoundError':
        return re.sub(r"No module named '(.*)'", r"没找到模块：‘\1’", 原信息)
    return 类型 + "：" + 原信息

def 类型中文化(类型):
    中英表 = {
        "NoneType": "空变量",
        "int": "整数变量",
        "bool": "真假变量",
        "function": "函数",
    }
    return 中英表[类型] if 类型 in 中英表 else 类型



from ast import *

'''
注释"研究"的待进一步揣摩
'''
比较操作符 = {
    In: 'in'
}

def 转源码(节点, 缩进量="  "):
    """
    本方法由语法树生成木兰源码，可用于实现 Python 到木兰源码的简单转换工具。
    """
    自述 = "/* 本文件由命令 `木兰 -兰 ` 自动生成. */\n"
    生成器 = 木兰生成器(缩进量, 头部=自述)
    生成器.visit(节点)
    return "".join(生成器.结果)

class 木兰生成器(NodeVisitor):

    def __init__(self, 缩进量, 头部=None):
        self.调试 = False
        self.结果 = []
        self.缩进量 = 缩进量
        self.缩进 = 0
        self.行数 = 0
        self.所有类型 = []
        if 头部 is not None:
            self.结果.append(头部)

    def 记录(self, 信息):
        if self.调试:
            print(信息)

    def 编写(self, 文本):
        if not isinstance(文本, str):
            raise AssertionError("未正确处理节点：" + repr(文本))
        if self.行数:
            if self.结果:
                self.结果.append('\n' * self.行数)
            self.结果.append(self.缩进量 * self.缩进)
            self.行数 = 0
        self.结果.append(文本)

    # 待加行号信息
    def 另起一行(self, 节点=None, 额外=0):
        self.行数 = max(self.行数, 1 + 额外)

    def 主体(self, 所有声明):
        self.编写(' {')
        self.缩进 += 1
        for 声明 in 所有声明:
            self.visit(声明)

        self.缩进 -= 1
        self.另起一行()
        self.编写('}')

    def 类型主体(self, 所有声明):
        self.编写(' {')
        self.缩进 += 1
        变量声明 = []

        for 声明 in 所有声明:
            if isinstance(声明, FunctionDef) or isinstance(声明, ClassDef):
                if 变量声明:
                    self.另起一行()
                    self.主体(变量声明)
                    变量声明 = []
                self.visit(声明)
            else:
                变量声明.append(声明)

        if 变量声明:
            self.另起一行()
            self.主体(变量声明)
        self.缩进 -= 1
        self.另起一行()
        self.编写('}')

    def 形参(self, 节点):
        # TODO: 避免重复
        需逗号 = []

        def 写逗号():
            if 需逗号:
                self.编写(', ')
            else:
                需逗号.append(True)

        空档 = [None] * (len(节点.args) - len(节点.defaults))
        for 形参, 默认值 in zip(节点.args, 空档 + 节点.defaults):
            写逗号()
            self.visit(形参)
            if 默认值 is not None:
                self.编写("=")
                self.visit(默认值)

        # 实际上木兰的变长形参并非用 * 声明，不知此段何用
        '''
        if 节点.vararg is not None:
            写逗号()
            self.编写('*' + 节点.vararg)
        if 节点.kwarg is not None:
            写逗号()
            self.编写("**" + 节点.kwarg)
        '''

    def visit_FunctionDef(self, 节点):
        self.另起一行(额外=1)
        self.另起一行(节点)

        取值 = False
        赋值 = False
        if 节点.decorator_list:

            # 研究：此处会报错，应为 节点.decorator_list
            for 修饰 in decorator_list:
                print(修饰)
                if isinstance(修饰, Name) and 修饰.id == 'property':
                    取值 = True
                elif isinstance(修饰, Attribute) and 修饰.attr == "setter":
                    赋值 = True
        if 取值:
            assert not 赋值
        if len(self.所有类型) > 0:
            if len(节点.args.args) > 0:
                if 节点.args.args[0].arg == 'self':
                    节点.args.args = 节点.args.args[1:]
                    if 节点.name == '__init__':
                        节点.name = self.所有类型[-1]
                    节点.name = '$' + 节点.name
        if 取值 or 赋值:
            self.编写('attr ')
        else:
            self.编写('func ')

        self.编写('%s%s(' % (节点.name, ' = ' if 赋值 else ''))
        self.visit(节点.args)
        self.编写(')')
        self.主体(节点.body)

    def visit_ClassDef(self, 节点):
        self.所有类型.append(节点.name)
        self.另起一行(额外=2)
        self.另起一行(节点)
        self.编写('type %s' % 节点.name)
        for 基类 in 节点.bases:
            if 基类 != 节点.bases[0]:
                self.编写(', ')
            else:
                self.编写(' : ')
            self.visit(基类)

        self.类型主体(节点.body)

        # 研究：为何不 :-1 ？
        self.所有类型 = self.所有类型[:-2]

    def visit_Attribute(self, 节点):
        if isinstance(节点.value, Name) and 节点.value.id == 'self':
            self.编写('$%s' % 节点.attr)
        elif isinstance(节点.value, Call) and isinstance(节点.value.func, Name) and 节点.value.func.id == 'super':
            self.编写('super')
            if 节点.attr != '__init__':
                self.编写('.' + 节点.attr)
        else:
            self.visit(节点.value)
            self.编写('.' + 节点.attr)

    def visit_Call(self, 节点):
        需逗号 = []

        def 写逗号():
            if 需逗号:
                self.编写(', ')
            else:
                需逗号.append(True)

        self.visit(节点.func)
        self.编写('(')
        for 实参 in 节点.args:
            写逗号()
            self.visit(实参)

        for 关键词 in 节点.keywords:
            写逗号()
            self.编写(关键词.arg + '=')
            self.visit(关键词.value)

        self.编写(')')

    def visit_Name(self, 节点):
        self.记录("Name: " + 节点.id)
        if 节点.id == 'print':
            self.编写('println')
        elif 节点.id == 'chr':
            self.编写('char')
        else:
            self.编写(节点.id)

    def visit_Str(self, 节点):
        self.编写(repr(节点.s))

    def visit_Num(self, 节点):
        self.编写(repr(节点.n))

    # 待补完
    def visit_Compare(self, 节点):
        self.编写('(')
        左边 = 节点.left
        已开头 = False
        for 操作符, 右边 in zip(节点.ops, 节点.comparators):
            if 已开头:
                self.编写(' and ')
            操作符 = 比较操作符[type(操作符)]
            if 'in' == 操作符:
                self.visit(右边)
                self.编写('.__contains__(')
                self.visit(左边)
                self.编写(')')
            左边 = 右边
            已开头 = True

        self.编写(')')

    def visit_Expr(self, 节点):
        self.记录("Expr: " + str(节点))
        self.另起一行()
        self.visit(节点.value)

    def visit_arguments(self, 节点):
        self.形参(节点)

    def visit_arg(self, 节点):
        self.编写(节点.arg)

def 转换(节点):
    return 转源码(节点)



class 动物:
  def __init__(self):
    print(1)

class 野生:
  def __init__(self, 数量):
    print(数量)

class 人(动物):
  def __init__(self):
    super().__init__()

class 狼(野生):
  def __init__(self):
    super().__init__(2)

人()
狼()


import imp
import math
import os
import sys
import time
import threading
from datetime import datetime
from pathlib import Path

from 木兰.分析器.语法分析器 import 语法分析器

# 参考：https://docs.python.org/3.7/library/threading.html#thread-objects
class 线程(threading.Thread):
    """
    跟踪线程包装器。
    """

    def __init__(自身, *args, **kw):
        threading.Thread.__init__(自身, *args, **kw)
        自身.已亡 = False

    def start(自身):
        自身._线程__运行_backup = 自身.run
        自身.run = 自身._线程__运行
        threading.Thread.start(自身)

    def __运行(自身):
        sys.settrace(自身.全局跟踪)
        自身._线程__运行_backup()
        自身.run = 自身._线程__运行_backup

    def 全局跟踪(自身, 栈, 事件, 参数):
        if 事件 == 'call':
            return 自身.本地跟踪
        return

    def 本地跟踪(自身, 栈, 事件, 参数):
        if 自身.已亡:
            if 事件 == 'line':
                raise SystemExit
        return 自身.本地跟踪

    def 杀死(自身):
        自身.已亡 = True


def 分析并编译(源码文件名):
    with open(源码文件名, 'r', encoding='utf-8') as 文件:
        源码 = 文件.read()

        分析器 = 语法分析器()
        节点 = 分析器.分析(源码=源码, 源码文件=源码文件名)
        return compile(节点, 源码文件名, 'exec')


def 加载木兰模块(名称, 全局, 源自=(), 目录相对层次=0):
    木兰源码路径 = str(Path(*(名称.split(".")))) + '.ul'
    可执行码 = 分析并编译(木兰源码路径)
    if 可执行码 is None:
        raise ModuleNotFoundError(名称)

    # TODO: 研究何用
    所有模块 = []
    后段 = 名称
    模块名 = lambda 模块名称: 所有模块[-1].__name__ + '.' + 模块名称 if 所有模块 else ""

    点位 = 0
    while 点位 != -1:
        点位 = 后段.find('.')
        前段, 后段 = 后段[:点位], 后段[点位 + 1:]
        if 点位 == -1:
            前段 = 后段
            模块 = imp.new_module(模块名(后段))
            模块.__dict__.update(创建全局变量(argv=(全局['ARGV'])))
            模块.__dict__['__file__'] = os.path.abspath(木兰源码路径)
            exec(可执行码, 模块.__dict__)
        else:
            模块 = imp.new_module(模块名(前段))
        if 所有模块:
            所有模块[-1].__dict__[前段] = 模块
        所有模块.append(模块)

    顶层 = 所有模块[0]
    if len(所有模块) > 1:
        if 源自 is not None:
            for sym in 源自:
                if sym == '*':
                    for k in 模块.__dict__:
                        if k not in 全局:
                            顶层.__dict__[k] = 模块.__dict__[k]
                    break
                else:
                    顶层.__dict__[sym] = 模块.__dict__[sym]

    return 顶层


def 内置扩展(内置项):
    from inspect import isclass
    for k, v in __builtins__.items():
        if isclass(v) and issubclass(v, BaseException):
            内置项[k] = v

    return 内置项


def __内置_除(a, b):
    if isinstance(a, int):
        if isinstance(b, int):
            return math.floor(a / b)
    return a / b


def __内置_求余(a, b):
    # TODO: 不解为何要在 a b 为整数时特殊处理
    return a % b


def 创建全局变量(argv=[], 文件名=''):
    def 转字符串(x):

        def 容器转为字符串(容器, 始='', 末=''):
            字符串 = 始
            for 序号, 该项 in enumerate(容器):
                if 序号:
                    字符串 += ', '
                字符串 += 转字符串(该项)

            字符串 += 末
            return 字符串

        if x is None:
            return 'nil'
        if isinstance(x, bool):
            return 'true' if x else 'false'
        if isinstance(x, list):
            return 容器转为字符串(x, '[', ']')
        if isinstance(x, tuple):
            return 容器转为字符串(x)
        return str(x)

    def 自定义输出(*各物件, 分隔符=' ', 终止符='', 文件=sys.stdout, flush=False):
        """打印内容到流, 默认到标准输出."""
        for 物件 in 各物件:
            文件.write(转字符串(物件))
            if 物件 != 各物件[(-1)]:
                文件.write(分隔符)

        文件.write(终止符)
        if flush:
            文件.flush()

    def 自定义导入(名称, 全局=None, 本地=None, 源自=(), 目录相对层次=0):
        """导入一个木兰模块, 如果没找到, 则导入 Python 模块
        """
        try:
            return 加载木兰模块(名称, 全局, 源自, 目录相对层次)
        except FileNotFoundError:
            return __import__(名称, 全局, 本地, 源自, 目录相对层次)
        except Exception as 例外:
            raise 例外

    def 本地断言(表达式, 反馈=None):
        assert 表达式, 反馈

    def 内置自身():
        """ 当前任务 ID """
        return threading.currentThread()

    def 生成新任务(任务名, *参数列表):

        # 待研究：各参数作用
        私有线程 = 线程(target=任务名, args=参数列表, daemon=True)
        私有线程.start()
        return 私有线程

    def 杀死任务(任务线程):
        if isinstance(任务线程, 线程):
            if 任务线程 == threading.currentThread():
                sys.exit()
            elif 任务线程.is_alive():
                任务线程.杀死()

    def 用pip安装(*包, 命令='install'):
        import pip._internal
        return pip._internal.main([命令, *包])

    def eval_print(expr):
        if expr is None:
            return
        try:
            expr()
        except Exception:
            自定义输出(expr, 终止符='\n')

    当前目录 = os.getcwd()
    if 当前目录 not in sys.path:
        sys.path.append(当前目录)
    return {
        'print': 自定义输出,
        'println': lambda *各物件: 自定义输出(*各物件, **{'终止符': '\n'}),
        'assert': 本地断言,
        'len': len,
        'enumerate': enumerate,
        'all': all,
        'any': any,
        'range': range,
        'round': round,
        'input': input,
        'reverse': reversed,
        'super': super,
        'locals': lambda: locals(),
        'bool': bool,
        'float': float,
        'int': int,
        'str': str,
        'list': list,
        'dict': dict,
        'set': set,
        'tuple': lambda *各实参: 各实参,
        'char': chr,
        'ord': ord,
        'bytes': lambda 文本, 编码='ascii': bytes(文本, encoding=编码),
        'typeof': lambda 对象: 对象.__class__.__name__,
        'isa': lambda x, 类型: isinstance(x, 类型),
        'max': max,
        'min': min,
        'map': map,
        'filter': filter,
        'zip': zip,
        'staticmethod': staticmethod,
        'property': property,
        'ceil': math.ceil,
        'floor': math.floor,
        'fabs': math.fabs,
        'sqrt': math.sqrt,
        'log': math.log,
        'log10': math.log10,
        'exp': math.exp,
        'pow': math.pow,
        'sin': math.sin,
        'cos': math.cos,
        'tan': math.tan,
        'asin': math.asin,
        'acos': math.acos,
        'atan': math.atan,
        'spawn': 生成新任务,
        'kill': 杀死任务,
        'self': 内置自身,
        '再会': sys.exit,
        'quit': sys.exit,
        'open': open,
        'install': 用pip安装,
        'time': time.time,
        'year': lambda: datetime.now().year,
        'month': lambda: datetime.now().month,
        'day': lambda: datetime.now().day,
        'hour': lambda: datetime.now().hour,
        'minute': lambda: datetime.now().minute,
        'second': lambda: datetime.now().second,
        'microsecond': lambda: datetime.now().microsecond,
        'sleep': time.sleep,
        'delay': lambda 毫秒数: time.sleep(毫秒数 / 1000),
        'delayMicroseconds': lambda 微秒数: time.sleep(微秒数 / 1000000),
        'PI': math.pi,
        'ARGV': argv,
        '__builtins__': 内置扩展({
            '__import__': 自定义导入,
            '__build_class__': __build_class__,
            '__name__': '__main__',
            '__file__': 文件名,
            '__print__': eval_print,
            '__div__': __内置_除,
            '__rem__': __内置_求余})
    }



import 图快

# TODO: 添加执行接口（输入木兰源码，返回输出内容）

# 参考： https://stackoverflow.com/questions/3906232/python-get-the-print-output-in-an-exec-statement
import sys
from io import StringIO
import contextlib

@contextlib.contextmanager
def stdoutIO(stdout=None):
    old = sys.stdout
    if stdout is None:
        stdout = StringIO()
    sys.stdout = stdout
    yield stdout
    sys.stdout = old

code = """
i = [0,1,2]
for j in i :
    print(j)
"""
def 运行():
    源码 = 代码文本框.获取(1.0, "end-1c")  # 获取当前选中的文本
    with stdoutIO() as s:
        try:
            exec(code)
        except:
            print("Something wrong with the code")
    输出文本框.插入(1.0, s.getvalue())

窗口 = 图快.主窗口类()
窗口.标题('木兰笔谈')
窗口.尺寸('800x600+100+100')

代码文本框 = 图快.文本框类(窗口, 高度=3)
代码文本框.常规布局()

按钮 = 图快.按钮类(窗口, 文本='运行', 宽度=15, 高度=2, 命令=运行)
按钮.常规布局()

输出文本框 = 图快.文本框类(窗口, 高度=3)
输出文本框.常规布局()

窗口.主循环()


class Person:
  def __init__(self):
    self.id = 1

  @property
  def a(self):
    return self.id

  @property
  def b(self):
    return 2

  @a.setter  # 此处 a, 下面 def 又是 a, 冗余
  def a(self, n):
    self.id += n

mulan = Person()
print(mulan.a + mulan.b)
mulan.a = 3
print(mulan.a)